import base64
from goblet.deploy import create_cloudfunction, destroy_cloudfunction

from goblet.config import GConfig
import logging

from goblet.handler import Handler
from goblet.client import get_default_project, get_default_location


log = logging.getLogger('goblet.deployer')
log.setLevel(logging.INFO)


class PubSub(Handler):
    def __init__(self, name, topics=None):
        self.name = name
        self.cloudfunction = f"projects/{get_default_project()}/locations/{get_default_location()}/functions/{name}"
        self.topics = topics or {}

    def register_topic(self, name, func, kwargs):
        topic = kwargs["topic"]
        kwargs = kwargs.pop('kwargs')
        attributes = kwargs.get("attributes", {})
        if self.topics.get(topic):
            self.topics[topic][name] = {
                "func": func,
                "attributes": attributes
            }
        else:
            self.topics[topic] = {
                name: {
                    "func": func,
                    "attributes": attributes
                }
            }

    def __call__(self, event, context):
        topic_name = context.resource.split('/')[-1]
        data = base64.b64decode(event['data']).decode('utf-8')
        attributes = event.get("attributes") or {}

        topic = self.topics.get(topic_name)
        if not topic:
            raise ValueError(f"Topic {topic_name} not found")

        # check attributes
        for name, info in topic.items():
            if info["attributes"].items() <= attributes.items():
                info["func"](data)
        return

    def __add__(self, other):
        self.topics.update(other.topics)
        return self

    def deploy(self, sourceUrl=None, entrypoint=None):
        if not self.topics:
            return

        log.info("deploying topic functions......")
        config = GConfig()
        user_configs = config.cloudfunction or {}
        for topic in self.topics:
            req_body = {
                "name": f"{self.cloudfunction}-topic-{topic}",
                "description": config.description or "created by goblet",
                "entryPoint": entrypoint,
                "sourceUploadUrl": sourceUrl,
                "eventTrigger": {
                    "eventType": "providers/cloud.pubsub/eventTypes/topic.publish",
                    "resource": f"projects/{get_default_project()}/topics/{topic}"
                },
                "runtime": config.runtime or "python37",
                **user_configs
            }
            create_cloudfunction(req_body)

    def destroy(self):
        for topic in self.topics:
            destroy_cloudfunction(f"{self.name}-topic-{topic}")



import json
from goblet.utils import get_g_dir
import os


class GConfig:
    def __init__(self, config=None, stage=None):
        self.config = config or self.get_g_config()
        self.stage = stage or os.environ.get("STAGE")
        self.validate()
        if self.stage:
            self.config.update(self.config.get("stages", {}).get(self.stage, {}))

    @staticmethod
    def get_g_config():
        try:
            with open(f'{get_g_dir()}/config.json') as f:
                return json.load(f)
        except FileNotFoundError:
            return {}

    def __getattr__(self, name):
        attr = self.config.get(name)
        if attr:
            return attr
        if os.environ.get(name):
            return os.environ.get(name)

        return None

    def __setattr__(self, name, value):
        if name not in ["config", "stage"]:
            self.config[name] = value
        else:
            super(GConfig, self).__setattr__(name, value)

    def write(self):
        with open(f'{get_g_dir()}/config.json', 'w') as f:
            f.write(json.dumps(self.config, indent=4))

    def validate(self):
        if self.stage and self.stage not in self.config.get("stages"):
            raise ValueError(f"stage {self.stage} not found in config")
        for stage in self.config.get("stages", {}):
            if "function_name" not in self.config["stages"][stage]:
                raise ValueError(f"function_name key missing for stage {stage}")



import logging

from goblet.handler import Handler
from goblet.client import Client, get_default_project, get_default_location
from googleapiclient.errors import HttpError

log = logging.getLogger('goblet.deployer')
log.setLevel(logging.INFO)


class Scheduler(Handler):
    def __init__(self, name, jobs=None):
        self.name = name
        self.cloudfunction = f"projects/{get_default_project()}/locations/{get_default_location()}/functions/{name}"
        self.jobs = jobs or {}
        self._api_client = None

    @property
    def api_client(self):
        if not self._api_client:
            self._api_client = self._create_api_client()
        return self._api_client

    def _create_api_client(self):
        return Client("cloudscheduler", 'v1', calls='projects.locations.jobs', parent_schema='projects/{project_id}/locations/{location_id}')

    def register_job(self, name, func, kwargs):
        schedule = kwargs["schedule"]
        kwargs = kwargs.pop('kwargs')
        timezone = kwargs.get("timezone", 'UTC')
        description = kwargs.get("description", "Created by goblet")
        self.jobs[name] = {
            "job_json": {
                "name": f"projects/{get_default_project()}/locations/{get_default_location()}/jobs/{self.name}-{name}",
                "schedule": schedule,
                "timeZone": timezone,
                "description": description,
                "httpTarget": {
                    # "uri": ADDED AT runtime,
                    "headers": {
                        'X-Goblet-Type': 'schedule',
                        'X-Goblet-Name': name
                    },
                    "httpMethod": "GET",
                    'oidcToken': {
                        # "serviceAccountEmail": ADDED AT runtime
                    }
                }
            },
            "func": func
        }

    def __call__(self, request, context=None):
        headers = request.headers
        func_name = headers.get("X-Goblet-Name")
        if not func_name:
            raise ValueError("No X-Goblet-Name header found")

        job = self.jobs[func_name]
        if not job:
            raise ValueError(f"Function {func_name} not found")
        return job["func"]()

    def __add__(self, other):
        self.jobs.update(other.jobs)
        return self

    def deploy(self, sourceUrl=None, entrypoint=None):
        if not self.jobs:
            return

        cloudfunction_client = Client("cloudfunctions", 'v1', calls='projects.locations.functions', parent_schema='projects/{project_id}/locations/{location_id}')
        resp = cloudfunction_client.execute('get', parent_key="name", parent_schema=self.cloudfunction)
        if not resp:
            raise ValueError(f"Function {self.cloudfunction} not found")
        cloudfunction_target = resp["httpsTrigger"]["url"]
        service_account = resp["serviceAccountEmail"]

        log.info("deploying scheduled jobs......")
        for job_name, job in self.jobs.items():
            job["job_json"]["httpTarget"]['uri'] = cloudfunction_target
            job["job_json"]["httpTarget"]['oidcToken']["serviceAccountEmail"] = service_account

            self.deploy_job(job_name, job["job_json"])

    def deploy_job(self, job_name, job):
        try:
            self.api_client.execute('create', params={'body': job})
            log.info(f"created scheduled job: {job_name} for {self.name}")
        except HttpError as e:
            if e.resp.status == 409:
                log.info(f"updated scheduled job: {job_name} for {self.name}")
                self.api_client.execute('patch', parent_key="name", parent_schema=job['name'], params={'body': job})
            else:
                raise e

    def destroy(self):
        if not self.jobs:
            return
        for job_name in self.jobs.keys():
            self._destroy_job(job_name)

    def _destroy_job(self, job_name):
        try:
            scheduler_client = Client("cloudscheduler", 'v1', calls='projects.locations.jobs', parent_schema='projects/{project_id}/locations/{location_id}/jobs/' + self.name + '-' + job_name)
            scheduler_client.execute('delete', parent_key="name")
            log.info("destroying scheduled functions......")
        except HttpError as e:
            if e.resp.status == 404:
                log.info("scheduled functions already destroyed")
            else:
                raise e



from collections import OrderedDict
from time import sleep
from marshmallow.schema import Schema
from ruamel import yaml
import base64
import logging
import re
from typing import get_type_hints
from apispec import APISpec
from apispec.ext.marshmallow import MarshmallowPlugin

from goblet.handler import Handler
from goblet.client import Client, get_default_project, get_default_location
from goblet.utils import get_g_dir
from goblet.config import GConfig
from googleapiclient.errors import HttpError

log = logging.getLogger('goblet.deployer')
log.setLevel(logging.INFO)


class ApiGateway(Handler):
    def __init__(self, app_name, routes=None):
        self.name = self.format_name(app_name)
        self.routes = routes or {}
        self._api_client = None
        self.cloudfunction = f"https://{get_default_location()}-{get_default_project()}.cloudfunctions.net/{self.name}"

    @property
    def api_client(self):
        if not self._api_client:
            self._api_client = self._create_api_client()
        return self._api_client

    def format_name(self, name):
        # ([a-z0-9-.]+) for api gateway name
        return name.replace('_', '-')

    def register_route(self, name, func, kwargs):
        path = kwargs.pop("path")
        methods = kwargs.pop("methods")
        kwargs = kwargs.pop('kwargs')
        path_entries = self.routes.get(path, {})
        for method in methods:
            if path_entries.get(method):
                raise ValueError(
                    "Duplicate method: '%s' detected for route: '%s'\n"
                    "between view functions: \"%s\" and \"%s\". A specific "
                    "method may only be specified once for "
                    "a particular path." % (
                        method, path, self.routes[path][method].function_name,
                        name)
                )
            entry = RouteEntry(func, name, path, method, **kwargs)
            path_entries[method] = entry
        self.routes[path] = path_entries

    def __call__(self, request, context=None):
        method = request.method
        path = request.path
        entry = self.routes.get(path, {}).get(method)
        if not entry:
            # test param paths
            for p in self.routes:
                if '{' in p and self._matched_path(p, path):
                    entry = self.routes.get(p, {}).get(method)
        # TODO: better handling
        if not entry:
            raise ValueError(f"No route found for {path} with {method}")
        return entry(request)

    def __add__(self, other):
        self.routes.update(other.routes)
        return self

    @staticmethod
    def _matched_path(org_path, path):
        split_org = re.sub(r'{[\w]+}', '', org_path).split('/')
        split_path = path.split('/')
        if len(split_path) != len(split_org):
            return False
        for i, item in enumerate(split_org):
            if item and split_path[i] != item:
                return False
        return True

    def _create_api_client(self):
        return Client("apigateway", 'v1', calls='projects.locations.apis', parent_schema='projects/{project_id}/locations/global')

    def _create_config_client(self):
        return Client("apigateway", 'v1', calls='projects.locations.apis.configs', parent_schema='projects/{project_id}/locations/global/apis/' + self.name)

    def _patch_config_client(self):
        return Client("apigateway", 'v1', calls='projects.locations.apis.configs', parent_schema='projects/{project_id}/locations/global/apis/' + self.name + '/configs/' + self.name)

    def _create_gateway_client(self):
        return Client("apigateway", 'v1', calls='projects.locations.gateways', parent_schema='projects/{project_id}/locations/{location_id}')

    def _patch_gateway_client(self):
        return Client("apigateway", 'v1', calls='projects.locations.gateways', parent_schema='projects/{project_id}/locations/{location_id}/gateways/' + self.name)

    def deploy(self, sourceUrl=None, entrypoint=None):
        if len(self.routes) == 0:
            return
        log.info("deploying api......")
        self.generate_openapi_spec(self.cloudfunction)
        try:
            resp = self.api_client.execute('create', params={'apiId': self.name})
            self.api_client.wait_for_operation(resp["name"])
        except HttpError as e:
            if e.resp.status == 409:
                log.info("api already deployed")
            else:
                raise e

        config = {
            "openapiDocuments": [
                {
                    "document": {
                        "path": f'{get_g_dir()}/{self.name}_openapi_spec.yml',
                        "contents": base64.b64encode(open(f'{get_g_dir()}/{self.name}_openapi_spec.yml', 'rb').read()).decode('utf-8')
                    }
                }
            ]
        }
        try:
            config_version_name = self.name
            self._create_config_client().execute('create', params={'apiConfigId': self.name, 'body': config})
        except HttpError as e:
            if e.resp.status == 409:
                log.info("updating api endpoints")
                configs = self._create_config_client().execute('list')
                # TODO: use hash
                version = len(configs['apiConfigs'])
                config_version_name = f"{self.name}-v{version}"
                self._create_config_client().execute('create', params={'apiConfigId': config_version_name, 'body': config})
            else:
                raise e
        gateway = {
            "apiConfig": f"projects/{get_default_project()}/locations/global/apis/{self.name}/configs/{config_version_name}",
        }
        try:
            gateway_resp = self._create_gateway_client().execute('create', params={'gatewayId': self.name, 'body': gateway})
        except HttpError as e:
            if e.resp.status == 409:
                log.info("updating gateway")
                gateway_resp = self._patch_gateway_client().execute('patch', parent_key='name', params={'updateMask': 'apiConfig', 'body': gateway})
            else:
                raise e
        if gateway_resp:
            self._create_gateway_client().wait_for_operation(gateway_resp["name"])
        log.info("api successfully deployed...")
        gateway_resp = self._patch_gateway_client().execute('get', parent_key='name')
        log.info(f"api endpoint is {gateway_resp['defaultHostname']}")
        return

    def destroy(self):
        if len(self.routes) == 0:
            return

        # destroy api gateway
        try:
            gateway_client = Client("apigateway", 'v1', calls='projects.locations.gateways', parent_schema='projects/{project_id}/locations/{location_id}/gateways/' + self.name)
            gateway_client.execute('delete', parent_key="name")
            log.info("destroying api gateway......")
        except HttpError as e:
            if e.resp.status == 404:
                log.info("api gateway already destroyed")
            else:
                raise e
        # destroy api config
        try:
            configs = self._create_config_client().execute('list')
            api_client = None
            resp = {}
            for c in configs.get('apiConfigs', []):
                api_client = Client("apigateway", 'v1', calls='projects.locations.apis.configs', parent_schema='projects/{project_id}/locations/global/apis/' + self.name + '/configs/' + c['displayName'])
                resp = api_client.execute('delete', parent_key="name")
            log.info("api configs destroying....")
            if api_client:
                api_client.wait_for_operation(resp["name"])
                sleep(2)
        except HttpError as e:
            if e.resp.status == 404:
                log.info("api configs already destroyed")
            else:
                raise e

        # destroy api
        try:
            api_client = Client("apigateway", 'v1', calls='projects.locations.apis', parent_schema='projects/{project_id}/locations/global/apis/' + self.name)
            api_client.execute('delete', parent_key="name")
            log.info("apis successfully destroyed......")
        except HttpError as e:
            if e.resp.status == 404:
                log.info("api already destroyed")
            else:
                raise e

    def generate_openapi_spec(self, cloudfunction):
        config = GConfig()
        spec = OpenApiSpec(self.name, cloudfunction, security_definitions=config.securityDefinitions)
        spec.add_apigateway_routes(self.routes)
        with open(f'{get_g_dir()}/{self.name}_openapi_spec.yml', 'w') as f:
            spec.write(f)


PRIMITIVE_MAPPINGS = {
    str: "string",
    bool: "boolean",
    int: "integer"
}


class OpenApiSpec:
    def __init__(self, app_name, cloudfunction, version="1.0.0", security_definitions=None):
        self.spec = OrderedDict()
        self.app_name = app_name
        self.cloudfunction = cloudfunction
        self.version = version
        self.spec["swagger"] = "2.0"
        self.spec["info"] = {
            "title": self.app_name,
            "description": "Goblet Autogenerated Spec",
            "version": self.version
        }
        if security_definitions:
            self.spec["securityDefinitions"] = security_definitions
        self.spec["schemes"] = ["https"]
        self.spec['produces'] = ["application/json"]
        self.spec["paths"] = {}
        self.component_spec = APISpec(
            title="",
            version="1.0.0",
            openapi_version="2.0",
            plugins=[MarshmallowPlugin()],
        )
        self.spec["definitions"] = {}

    def add_component(self, component):
        if component.__name__ in self.component_spec.components.schemas:
            return
        self.component_spec.components.schema(component.__name__, schema=component)
        self.spec["definitions"] = self.component_spec.to_dict()['definitions']

    def add_apigateway_routes(self, apigateway):
        for path, methods in apigateway.items():
            for method, entry in methods.items():
                self.add_route(entry)

    def get_param_type(self, type_info, only_primititves=False):
        if not type_info:
            return {"type": "string"}
        if type_info in PRIMITIVE_MAPPINGS.keys():
            param_type = {"type": PRIMITIVE_MAPPINGS[type_info]}
        elif issubclass(type_info, Schema) and not only_primititves:
            self.add_component(type_info)
            param_type = {"$ref": f"#/definitions/{type_info.__name__}"}
        else:
            raise ValueError(f"param_type has type {type_info}. \
                It must be of type {PRIMITIVE_MAPPINGS.values} or a dataclass inheriting from Schema")
        return param_type

    def add_route(self, entry):
        method_spec = OrderedDict()
        method_spec["x-google-backend"] = {
            "address": self.cloudfunction,
            "protocol": "h2",
            "path_translation": "APPEND_PATH_TO_ADDRESS"
        }
        method_spec["operationId"] = entry.function_name

        params = []
        type_hints = get_type_hints(entry.route_function)
        for param in entry.view_args:
            type_info = type_hints.get(param, str)
            param_type = self.get_param_type(type_info, only_primititves=True)

            param_entry = {
                "in": "path",
                "name": param,
                "required": True,
                **param_type
            }
            params.append(param_entry)
        if params:
            method_spec["parameters"] = params
        if entry.request_body:
            if isinstance(entry.request_body, dict):
                method_spec["requestBody"] = entry.request_body

        # TODO: add query strings

        return_type = type_hints.get('return')
        content = {}
        if return_type:
            if return_type in PRIMITIVE_MAPPINGS.keys():
                content = {
                    "schema": {
                        "type": PRIMITIVE_MAPPINGS[return_type]
                    }
                }
            # list
            elif "typing.List" in str(return_type):
                type_info = return_type.__args__[0]
                param_type = self.get_param_type(type_info)
                content = {
                    "schema": {
                        "type": "array",
                        "items": {
                            **param_type
                        }
                    }
                }
            elif issubclass(return_type, Schema):
                param_type = self.get_param_type(return_type)
                content = {
                    "schema": {
                        **param_type
                    }
                }
        if entry.responses:
            method_spec["responses"] = entry.responses
        else:
            method_spec["responses"] = {
                '200': {
                    "description": "A successful response",
                    **content
                }
            }
        path_exists = self.spec["paths"].get(entry.uri_pattern)
        if path_exists:
            self.spec["paths"][entry.uri_pattern][entry.method.lower()] = dict(method_spec)
        else:
            self.spec["paths"][entry.uri_pattern] = {
                entry.method.lower(): dict(method_spec)
            }

    def write(self, file):
        yaml.Representer.add_representer(OrderedDict, yaml.Representer.represent_dict)
        yaml.YAML().dump(dict(self.spec), file)


_PARAMS = re.compile(r'{\w+}')


class RouteEntry:

    def __init__(self, route_function, function_name, path, method,
                 api_key_required=None, content_types=None,
                 cors=False, **kwargs):
        self.route_function = route_function
        self.function_name = function_name
        self.uri_pattern = path
        self.method = method
        self.api_key_required = api_key_required
        self.request_body = kwargs.get("request_body")
        self.responses = kwargs.get("responses")
        #: A list of names to extract from path:
        #: e.g, '/foo/{bar}/{baz}/qux -> ['bar', 'baz']
        self.view_args = self._parse_view_args()
        self.content_types = content_types
        # cors is passed as either a boolean or a CORSConfig object. If it is a
        # boolean it needs to be replaced with a real CORSConfig object to
        # pass the typechecker. None in this context will not inject any cors
        # headers, otherwise the CORSConfig object will determine which
        # headers are injected.
        if cors is True:
            cors = CORSConfig()
        elif cors is False:
            cors = None
        self.cors = cors
        self.kwargs = {**kwargs}

    def _extract_view_args(self, path):
        components = path.split('/')
        original_components = self.uri_pattern.split('/')
        matches = 0
        args = {}
        for i, component in enumerate(components):
            if component != original_components[i]:
                args[self.view_args[matches]] = component
                matches += 1
        return args

    def __call__(self, request):
        # TODO: pass in args and kwargs and options
        args = self._extract_view_args(request.path)
        return self.route_function(**args)

    def _parse_view_args(self):
        if '{' not in self.uri_pattern:
            return []
        # The [1:-1] slice is to remove the braces
        # e.g {foobar} -> foobar
        results = [r[1:-1] for r in _PARAMS.findall(self.uri_pattern)]
        return results

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class CORSConfig(object):
    """A cors configuration to attach to a route."""

    _REQUIRED_HEADERS = ['Content-Type', 'Authorization']

    def __init__(self, allow_origin='*', allow_headers=None,
                 expose_headers=None, max_age=None, allow_credentials=None):
        self.allow_origin = allow_origin

        if allow_headers is None:
            allow_headers = set(self._REQUIRED_HEADERS)
        else:
            allow_headers = set(allow_headers + self._REQUIRED_HEADERS)
        self._allow_headers = allow_headers

        if expose_headers is None:
            expose_headers = []
        self._expose_headers = expose_headers

        self._max_age = max_age
        self._allow_credentials = allow_credentials

    @property
    def allow_headers(self):
        return ','.join(sorted(self._allow_headers))

    def get_access_control_headers(self):
        headers = {
            'Access-Control-Allow-Origin': self.allow_origin,
            'Access-Control-Allow-Headers': self.allow_headers
        }
        if self._expose_headers:
            headers.update({
                'Access-Control-Expose-Headers': ','.join(self._expose_headers)
            })
        if self._max_age is not None:
            headers.update({
                'Access-Control-Max-Age': str(self._max_age)
            })
        if self._allow_credentials is True:
            headers.update({
                'Access-Control-Allow-Credentials': 'true'
            })

        return headers

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.get_access_control_headers() == \
                other.get_access_control_headers()
        return False



import os
import importlib.util
from contextlib import contextmanager


@contextmanager
def add_to_path(p):
    import sys
    old_path = sys.path
    sys.path = sys.path[:]
    sys.path.insert(0, p)
    try:
        yield
    finally:
        sys.path = old_path


def get_app_from_module(m):
    from goblet import Goblet
    for obj in dir(m):
        if isinstance(getattr(m, obj), Goblet):
            return getattr(m, obj), obj


def get_goblet_app():
    # looks for main.py and gets goblet app
    dir_path = os.path.realpath('.')
    spec = importlib.util.spec_from_file_location("main", f"{dir_path}/main.py")
    main = importlib.util.module_from_spec(spec)
    with add_to_path(dir_path):
        spec.loader.exec_module(main)
        app, app_name = get_app_from_module(main)
    setattr(app, "entrypoint", app_name)
    return app


def get_g_dir():
    return f"{os.path.realpath('.')}/.goblet"


def get_dir():
    return os.path.realpath('.')



from goblet.config import GConfig
import click
import os
import logging
import subprocess

from goblet.utils import get_goblet_app
from goblet.deploy import Deployer
from goblet.client import get_default_project
from goblet.__version__ import __version__

logging.basicConfig()


@click.group()
def main():
    pass


@main.command()
def help():
    click.echo('Use goblet --help. You can also view the full docs for goblet at https://anovis.github.io/goblet/docs/build/html/index.html')


@main.command()
def version():
    """current goblet version"""
    click.echo(__version__)


@main.command()
@click.option('-p', '--project', 'project', envvar='GOOGLE_PROJECT')
@click.option('-l', '--location', 'location', envvar='GOOGLE_LOCATION', required=True)
@click.option('-s', '--stage', 'stage', envvar='STAGE')
@click.option('--skip-function', 'skip_function', is_flag=True)
@click.option('--only-function', 'only_function', is_flag=True)
def deploy(project, location, stage, skip_function, only_function):
    """
    You can set the project and location using environment variable GOOGLE_PROJECT and GOOGLE_LOCATION

    Note: Allowed GOOGLE_LOCATION values for API GATEWAY are: asia-east1, europe-west1, us-eastl1 and us-central1.

    Note: Make sure api-gateway, cloudfunctions, and storage are enabled in your project
    """
    try:
        _project = project or get_default_project()
        if not _project:
            click.echo("Project not found. Set --project flag or add to gcloud by using gcloud config set project PROJECT")
        os.environ["GOOGLE_PROJECT"] = _project
        os.environ["GOOGLE_LOCATION"] = location
        if stage:
            os.environ["STAGE"] = stage
        app = get_goblet_app()
        Deployer({"name": app.function_name}).deploy(app, skip_function=skip_function, only_function=only_function)

    except FileNotFoundError:
        click.echo("Missing main.py. This is the required entrypoint for google cloud functions")


@main.command()
@click.option('-p', '--project', 'project', envvar='GOOGLE_PROJECT')
@click.option('-l', '--location', 'location', envvar='GOOGLE_LOCATION', required=True)
@click.option('-s', '--stage', 'stage', envvar='STAGE')
def destroy(project, location, stage):
    """
    Deletes all resources in gcp that are defined the current deployment
    """
    try:
        _project = project or get_default_project()
        if not _project:
            click.echo("Project not found. Set --project flag or add to gcloud by using gcloud config set project PROJECT")
        os.environ["GOOGLE_PROJECT"] = _project
        os.environ["GOOGLE_LOCATION"] = location
        if stage:
            os.environ["STAGE"] = stage
        app = get_goblet_app()
        Deployer({"name": app.function_name}).destroy(app)

    except FileNotFoundError:
        click.echo("Missing main.py. This is the required entrypoint for google cloud functions")


@main.command()
@click.argument('cloudfunction')
def openapi(cloudfunction):
    """
    You can find the generated openapi spec in /.goblet folder.

    The cloudfunction argument sets the correct x-google-backend address in the openapi spec.
    """
    try:
        app = get_goblet_app()
        app.handlers["route"].generate_openapi_spec(cloudfunction)

    except FileNotFoundError:
        click.echo("Missing main.py. This is the required entrypoint for google cloud functions")


@main.command()
@click.argument('local_arg',)
def local(local_arg):
    """
    Requires the local argument to be set in the Goblet class.

    For example in this case you would use local_function

    Goblet("test_function",local="local_function")
    """
    try:
        subprocess.check_output(["functions-framework", f"--target={local_arg}", "--debug"])
    except subprocess.CalledProcessError:
        click.echo("Incorrect argument. Make sure you set the local param in your Goblet class and that it matches the arg used in goblet local")


@click.option('-s', '--stage', 'stage', envvar='STAGE')
@main.command()
def package(stage):
    """generates the goblet zipped package in .goblet folder"""
    try:
        if stage:
            os.environ["STAGE"] = stage
        app = get_goblet_app()
        Deployer({"name": app.function_name}).package()

    except FileNotFoundError:
        click.echo("Missing main.py. This is the required entrypoint for google cloud functions")


@main.group()
def stage():
    """view and create different environment stages"""
    pass


@stage.command(name="list")
def list_stages():
    config = GConfig.get_g_config()
    if not config.get("stages"):
        return click.echo("no stages found")
    click.echo(list(config["stages"].keys()))


@stage.command()
@click.argument('stage',)
def create(stage):
    """create a new stage in config.json"""
    config = GConfig()
    if config.stages and stage in config.stages:
        return click.echo(f"stage {stage} already exists")
    app = get_goblet_app()
    function_name = f"{app.function_name}-{stage}"
    if not config.stages:
        config.stages = {stage: {"function_name": function_name}}
    else:
        config.stages[stage] = {"function_name": function_name}
    config.write()
    click.echo(f"stage {stage} created in config.json with function name {function_name}")



import os
import time

from googleapiclient.discovery import build

import google.auth


def get_default_project():
    for k in ('GOOGLE_PROJECT', 'GCLOUD_PROJECT',
              'GOOGLE_CLOUD_PROJECT', 'CLOUDSDK_CORE_PROJECT'):
        if k in os.environ:
            return os.environ[k]
        try:
            _, project = google.auth.default()
            return project
        except Exception:
            return None

    return None


def get_default_location():
    for k in ('GOOGLE_ZONE', 'GCLOUD_ZONE', 'CLOUDSDK_COMPUTE_ZONE',
              'GOOGLE_REGION', 'GCLOUD_REGION', 'CLOUDSDK_COMPUTE_REGION',
              'GOOGLE_LOCATION', 'GCLOUD_LOCATION'):
        if k in os.environ:
            return os.environ[k]

    return None


def get_credentials():
    """get user credentials and save them for future use
    """
    credentials, project = google.auth.default()
    return credentials


class Client:
    def __init__(self, resource, version='v1', credentials=None, calls=None, parent_schema=None):
        self.project_id = get_default_project()
        self.location_id = get_default_location()
        self.calls = calls
        self.resource = resource
        self.version = version
        self.parent_schema = parent_schema
        self.credentials = credentials or get_credentials()

        self.client = build(resource, version, credentials=self.credentials, cache_discovery=False)

        self.parent = None
        if self.parent_schema:
            self.parent = self.parent_schema.format(project_id=self.project_id, location_id=self.location_id)

    def __call__(self):
        return self.client

    def wait_for_operation(self, operation, timeout=600, calls="projects.locations.operations"):
        done = False
        operation_client = Client(
            self.resource,
            version=self.version,
            credentials=self.credentials,
            calls=calls,
            parent_schema=operation
        )
        count = 0
        sleep_duration = 4
        while not done or count > timeout:
            resp = operation_client.execute('get', parent_key="name")
            done = resp.get("done")
            time.sleep(sleep_duration)
            count += sleep_duration

    def execute(self, api, calls=None, parent_schema=None, parent=True, parent_key='parent', params=None):
        api_chain = self.client
        _params = params or {}
        _calls = calls or self.calls
        if parent_schema:
            parent_schema = parent_schema.format(project_id=self.project_id, location_id=self.location_id)
        _schema = parent_schema or self.parent

        if isinstance(_calls, str):
            calls = _calls.split('.')
        for call in calls:
            api_chain = getattr(api_chain, call)()

        if _schema and parent:
            _params[parent_key] = _schema
        return getattr(api_chain, api)(**_params).execute()



from pathlib import Path
import zipfile
import os
import requests
import logging
import hashlib

from googleapiclient.errors import HttpError

from goblet.client import Client, get_default_project, get_default_location
from goblet.utils import get_dir, get_g_dir
from goblet.config import GConfig

log = logging.getLogger('goblet.deployer')
log.setLevel(logging.INFO)


class Deployer:

    def __init__(self, config={}):
        self.config = config
        if not config:
            self.config = {
                "name": "goblet_test_app"
            }
        self.name = self.config["name"]
        self.zipf = self.create_zip()
        self.goblet_hash_name = self.project_hash()
        self.function_client = self._create_function_client()

    def _create_function_client(self):
        return Client("cloudfunctions", 'v1', calls='projects.locations.functions', parent_schema='projects/{project_id}/locations/{location_id}')

    def project_hash(self):
        if not get_default_project():
            return None
        m = hashlib.md5()
        m.update(get_default_project().encode('utf-8'))
        m.update(self.config["name"].encode('utf-8'))
        m.update(b"goblet")
        return "goblet-" + m.hexdigest()

    def package(self):
        self.zip()

    def deploy(self, goblet, skip_function=False, only_function=False, config=None):
        url = None
        if not skip_function:
            log.info("zipping function......")
            self.zip()
            log.info("uploading function zip to gs......")
            url = self._upload_zip()
            # TODO: CHECK IF VERSION IS DEPLOYED
            if goblet.is_http():
                self.create_function(url, goblet.entrypoint)
        if not only_function:
            goblet.deploy(url)

        return goblet

    def destroy(self, goblet):
        goblet.destroy()
        destroy_cloudfunction(self.name)

        return goblet

    def create_function(self, url, entrypoint):
        config = GConfig()
        user_configs = config.cloudfunction or {}
        req_body = {
            "name": f"projects/{get_default_project()}/locations/{get_default_location()}/functions/{self.name}",
            "description": config.description or "created by goblet",
            "entryPoint": entrypoint,
            "sourceUploadUrl": url,
            "httpsTrigger": {},
            "runtime": "python37",
            **user_configs
        }
        create_cloudfunction(req_body)

    def _upload_zip(self):
        self.zipf.close()
        zip_size = os.stat(f'.goblet/{self.name}.zip').st_size
        with open(f'.goblet/{self.name}.zip', 'rb') as f:
            resp = self.function_client.execute('generateUploadUrl')

            requests.put(
                resp["uploadUrl"],
                data=f,
                headers={
                    "content-type": "application/zip",
                    'Content-Length': str(zip_size),
                    "x-goog-content-length-range": "0,104857600"
                }
            )

        log.info("function code uploaded")

        return resp["uploadUrl"]

    def create_zip(self):
        if not os.path.isdir(get_g_dir()):
            os.mkdir(get_g_dir())
        return zipfile.ZipFile(get_g_dir() + f'/{self.name}.zip', 'w', zipfile.ZIP_DEFLATED)

    def zip(self):
        config = GConfig()
        self.zip_file("requirements.txt")
        include = config.customFiles or []
        include.append('*.py')
        self.zip_directory(get_dir() + '/*', include=include)

    def zip_file(self, filename):
        self.zipf.write(filename)

    def zip_directory(self, dir, include=['*.py'], exclude=['build', 'docs', 'examples']):
        exclusion_set = set(exclude)
        globbed_files = []
        for pattern in include:
            globbed_files.extend(Path('').rglob(pattern))
        for path in globbed_files:
            if not set(path.parts).intersection(exclusion_set):
                self.zipf.write(str(path))


def create_cloudfunction(req_body):
    function_client = Client("cloudfunctions", 'v1', calls='projects.locations.functions', parent_schema='projects/{project_id}/locations/{location_id}')
    try:
        resp = function_client.execute('create', parent_key="location", params={'body': req_body})
        log.info(f"creating cloudfunction {req_body['name']}")
    except HttpError as e:
        if e.resp.status == 409:
            log.info(f"updating cloudfunction {req_body['name']}")
            resp = function_client.execute('patch', parent_key="name", parent_schema=req_body["name"], params={'body': req_body})
        else:
            raise e
    function_client.wait_for_operation(resp["name"], calls="operations")


def destroy_cloudfunction(name):
    try:
        client = Client("cloudfunctions", 'v1', calls='projects.locations.functions', parent_schema='projects/{project_id}/locations/{location_id}/functions/' + name)
        client.execute('delete', parent_key="name")
        log.info(f"deleting google cloudfunction {name}......")
    except HttpError as e:
        if e.resp.status == 404:
            log.info(f"cloudfunction {name} already destroyed")
        else:
            raise e



from goblet.resources.pubsub import PubSub
from goblet.resources.routes import ApiGateway
from goblet.resources.scheduler import Scheduler
import logging

log = logging.getLogger(__name__)

EVENT_TYPES = ["all", 'http', 'schedule', 'pubsub']


class DecoratorAPI:

    def middleware(self, event_type='all'):
        if event_type not in EVENT_TYPES:
            raise ValueError(f"{event_type} not in {EVENT_TYPES}")

        def _middleware_wrapper(func):
            self.register_middleware(func, event_type)
            return func
        return _middleware_wrapper

    def route(self, path, methods=['GET'], **kwargs):
        return self._create_registration_function(
            handler_type='route',
            registration_kwargs={'path': path, 'methods': methods, 'kwargs': kwargs},
        )

    def schedule(self, schedule, **kwargs):
        return self._create_registration_function(
            handler_type='schedule',
            registration_kwargs={'schedule': schedule, 'kwargs': kwargs},
        )

    def topic(self, topic, **kwargs):
        return self._create_registration_function(
            handler_type='pubsub',
            registration_kwargs={'topic': topic, 'kwargs': kwargs},
        )

    def _create_registration_function(self, handler_type,
                                      registration_kwargs=None):
        def _register_handler(user_handler):
            handler_name = user_handler.__name__
            kwargs = registration_kwargs or {}
            self._register_handler(handler_type, handler_name,
                                   user_handler, kwargs)
            return user_handler
        return _register_handler

    def _register_handler(self, handler_type, name,
                          func, kwargs, options=None):
        raise NotImplementedError("_register_handler")

    def register_middleware(self, func, event_type='all'):
        raise NotImplementedError("register_middleware")


class Register_Handlers(DecoratorAPI):

    def __init__(self, function_name):
        self.handlers = {
            "route": ApiGateway(function_name),
            "schedule": Scheduler(function_name),
            "pubsub": PubSub(function_name)
        }
        self.middleware_handlers = {}
        self.current_request = None

    def __call__(self, request, context=None):
        self.current_request = request
        event_type = self.get_event_type(request, context)

        # call middleware
        request = self._call_middleware(request, event_type)

        if event_type == "schedule":
            return self.handlers['schedule'](request)
        if event_type == "pubsub":
            return self.handlers['pubsub'](request, context)
        if event_type == "http":
            self.handlers["route"](request)

        raise ValueError(f"{event_type} not a valid event type")

    def __add__(self, other):
        for handler in self.handlers:
            self.handlers[handler] += other.handlers[handler]
        return self

    def combine(self, other):
        return self + other

    def get_event_type(self, request, context=None):
        if context and context.event_type:
            return context.event_type.split('.')[1].split('/')[0]
        if request.headers.get("X-Goblet-Type") == 'schedule':
            return "schedule"
        return 'http'

    def _call_middleware(self, event, event_type):
        middleware = self.middleware_handlers.get('all', [])
        middleware.extend(self.middleware_handlers.get(event_type, []))
        for m in middleware:
            event = m(event)

        return event

    def _register_handler(self, handler_type, name,
                          func, kwargs, options=None):

        getattr(self, '_register_%s' % handler_type)(
            name=name,
            func=func,
            kwargs=kwargs,
        )

    def deploy(self, sourceUrl):
        for k, v in self.handlers.items():
            log.info(f"deploying {k}")
            v.deploy(sourceUrl, self.entrypoint)

    def destroy(self):
        for k, v in self.handlers.items():
            log.info(f"deploying {k}")
            v.destroy()

    def is_http(self):
        if len(self.handlers["route"].routes) > 0 or len(self.handlers["schedule"].jobs) > 0:
            return True
        return False

    def register_middleware(self, func, event_type='all'):
        middleware_list = self.middleware_handlers.get(event_type, [])
        middleware_list.append(func)
        self.middleware_handlers[event_type] = middleware_list

    def _register_route(self, name, func, kwargs):
        self.handlers["route"].register_route(name=name, func=func, kwargs=kwargs)

    def _register_schedule(self, name, func, kwargs):
        self.handlers["schedule"].register_job(name=name, func=func, kwargs=kwargs)

    def _register_pubsub(self, name, func, kwargs):
        self.handlers["pubsub"].register_topic(name=name, func=func, kwargs=kwargs)
